package isac.wsdialog;

import isac.bridge.*;
import isac.session.SDialog;
import isac.util.*;
import isac.util.formulae.*;
import isac.util.interfaces.*;
import isac.util.tactics.*;
import isac.util.users.IUserSettings;
import isac.util.users.User;
import isac.util.users.UserSettings;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.rmi.*;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * The DialogGuide moderates the communication between two instances working on
 * the same CalcTree object. One of these instances is the user, the other a
 * math engine. Most probably, the user's GUI, the DialogGuide and the math
 * engine reside on different machines and communicate with each other by means
 * of RMI.
 * 
 * @author Alan Krempler
 */
public class DialogGuide extends UnicastRemoteObject implements DGuide,
		IToUser, Serializable {

	static Logger logger_ = Logger.getLogger(BridgeRMI.class.getName());

	protected MathEngine math_engine_;

	//the DialogGuide knowing the CalcTree is the price for closing off
	//the CalcTree from the Worksheet (before DGuide implemented IToCalc!)
	protected CalcTree calc_tree_;

	//protected CalcHead calc_head_;

	//protected CalcHead calc_head_sub_;

	protected int phase_;

	private Vector datachange_listeners_;

	private IToWorksheet ui_control_listener_;

	//protected String session_id_;WN0503

	protected UserSettings user_settings_;//hold whole hashmap locally

	/**
	 * @param ME_path
	 *            URL of the math engine to be used
	 * @param user_path
	 *            path(without URL) to UserSettings, UserModel
	 * @param session_dialog
	 * @param session_id
	 *            which create the DialogGuide and Worksheet
	 * @throws RemoteException
	 */
	public DialogGuide(String ME_path, String user_path,
			SDialog session_dialog, String session_id) throws RemoteException {
		super();
		this.rmiBind();
		datachange_listeners_ = new Vector();
		MathEngine.init(ME_path);
		math_engine_ = MathEngine.getMathEngine();
		phase_ = DIALOGPHASE_IDLE;

		User user = null;
		try {
			user = session_dialog.getUser(session_id);
		} catch (RemoteException e1) {
			e1.printStackTrace();
		}
		user_settings_ = new UserSettings(user_path, user.getUsername());
	}

	/**
	 * start a calculation
	 */
	public CalcTree startCalculation(int user_id, Formalization f,
			int start_from/* WN050421 , int requested_calchead_view */) {

		phase_ = DIALOGPHASE_SOLVE;
		calc_tree_ = math_engine_.startCalculation(f);
		calc_tree_.addDataChangeListener(this);
		return calc_tree_;
	}

	/**
	 * at the beginning of a calculation as well as of a subproblem;
	 * 
	 * @return CalcHead eg. for providing CalcHead#VIEWSTYLE_*
	 * 
	 * @see isac.wsdialog.DGuide#startSpecifying()
	 */
	public CalcHead startSpecifying() {
		ICalcIterator it = null;
		CalcHead calc_head = null;
		int calchead_view;
		try {
			it = calc_tree_.iterator();
			//			WN0504DI it = this.iterator();//WN050421 NEW...
			it.moveRoot();
			//WN0504DI ...error DialogGuide_Stub.startSpecifying(Unknown
			//WN0504DI Source) with ...NEW
			calc_head = (CalcHead) it.getFormula();
		} catch (RemoteException e) {
			e.printStackTrace();
		}

		//AK set viewstyle according to user settings
		switch (user_settings_.getValue(IUserSettings.KEY1_START_SPECIFY)) {
		//________________________________________________________
		case IUserSettings.VAL1_SKIP_SPECIFY_TO_START_SOLVE:
			calc_tree_.completeCalcHead();
			calchead_view = CalcHead.VIEWSTYLE_SINGLELINE;
			phase_ = DIALOGPHASE_SOLVE;
			break;
		//________________________________________________________
		case IUserSettings.VAL1_POP_CALCHEAD_WITH_DESCRIPTIONS:
			calc_tree_.autoCalculate(IToCalc.SCOPE_CALCULATION, 1);
			//...applies the Tactic Model_Problem which generates descriptions
			calchead_view = CalcHead.VIEWSTYLE_FULL;
			phase_ = DIALOGPHASE_SPECIFY;
			break;
		//________________________________________________________
		default:
			//=case IUserSettings.VAL1_SKIP_SPECIFY_TO_START_SOLVE:
			calc_tree_.completeCalcHead();
			calchead_view = CalcHead.VIEWSTYLE_SINGLELINE;
			phase_ = DIALOGPHASE_SOLVE;
			break;
		}
		calc_head.setViewStyle(calchead_view);
		return calc_head;
	}

	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public void modifyCalcHead(CalcHead calc_head) {
	//		// MathEngine.getMathEngine().modifyCalcHead(calc_head);
	//		calc_tree_.modifyCalcHead(calc_head);
	//	}

	/**
	 * dismiss the CalcHeadPanel and start DGuide#DIALOGPHASE_SOLVE
	 * 
	 * @see isac.util.interfaces.IToCalc#startSolving()
	 */
	public void startSolving() throws Exception {
		calc_tree_.startSolving();

	}

	private CalcChangedEvent toDialogIterators(CalcChangedEvent event) {
		event.getLastUnchangedFormula();
		return event;
	}

	/**
	 * accept the changes in a calculation reported by the bridge ..............
	 * TODO handle these changes according to UserSettings and UserModel
	 * 
	 * ... and pass them to the WorkSheet
	 * 
	 * @see isac.wsdialog.IToUser#calcChanged(isac.wsdialog.CalcChangedEvent)
	 */
	public void calcChanged(CalcChangedEvent event) throws RemoteException {
		logger_.fatal(" WS<-. . : calcChanged (unc="
				+ event.getLastUnchangedFormula().toSMLString() + ", del="
				+ event.getLastDeletedFormula().toSMLString() + ", gen="
				+ event.getLastGeneratedFormula().toSMLString());
		try {
			for (int i = 0; i < datachange_listeners_.size(); i++) {
				((IToUser) datachange_listeners_.elementAt(i))
						.calcChanged(event);
				//WN050511->AK?LK? ^^^^^^^^^ cast to DialogIterators ?
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * the <code>Dialog#iterator</code> calls methods upon the
	 * <code>CalcTree#iterator</code> eg. in <code>getFormula</code>.
	 * 
	 * @param CalcIterator
	 *            which causes the construction
	 * 
	 * @see isac.wsdialog.IToCalc#iterator()
	 */
	public ICalcIterator iterator(ICalcIterator it) {
		//WN050512
		//		try {
		//			return new DialogIterator((CalcIterator) calc_tree_.iterator(),
		//			//.......................(it, this);
		//					this);
		//		} catch (Exception e) {
		//			e.printStackTrace();
		//		}
		//		return null;
		try {
			return new DialogIterator((CalcIterator) it, (DGuide) this);
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @see isac.wsdialog.IToCalc#addListener(isac.wsdialog.IToUser)
	 */
	public boolean addDataChangeListener(IToUser listener)
			throws RemoteException {
		if (datachange_listeners_.contains(listener)) {
			return false;
		} else {
			datachange_listeners_.add(listener);
		}
		return true;
	}

	public boolean registerUIControlListener(IToWorksheet listener)
			throws RemoteException {
		//WN0504 ???
		//		if (ui_control_listener_ != null) {
		//			ui_control_listener_.doUIAction(new UserAction(UI_DO_DETACH));
		//		}
		ui_control_listener_ = listener;
		return true;
	}

	//	/**
	//	 * @see isac.wsdialog.IToCalc#fetchProposedTactic
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public Tactic fetchProposedTactic() {
	//		return calc_tree_.fetchProposedTactic();
	//	}
	//
	//
	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public int setNextTactic(Tactic tactic) {
	//		return calc_tree_.setNextTactic(tactic);
	//	}
	//
	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public int autoCalculate(int scope, int nSteps) {
	//		return calc_tree_.autoCalculate(scope, nSteps);
	//	}
	//
	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public void tryMatch(CalcHead ch, CalcHeadCompoundID problemID)
	//			throws NotInSpecificationPhaseException {
	//
	//		// MathEngine.getMathEngine().tryMatch(ch, problemID);
	//		math_engine_.tryMatch(ch, problemID);
	//	}
	//
	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public void tryRefine(CalcHead ch, CalcHeadCompoundID problemID)
	//			throws NotInSpecificationPhaseException {
	//
	//		// MathEngine.getMathEngine().tryRefine(ch , problemID);
	//		math_engine_.tryRefine(ch, problemID);
	//	}
	//
	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public ICalcIterator getActiveFormula() {
	//		return calc_tree_.getActiveFormula();
	//	}
	//
	//	/**
	//	 * @throws RemoteException
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public int replaceFormula(CalcFormula newFormula) throws RemoteException
	// {
	//		return calc_tree_.replaceFormula(newFormula);
	//	}
	//
	//	/**
	//	 * @throws RemoteException
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public int appendFormula(CalcFormula newFormula) throws RemoteException {
	//		return calc_tree_.appendFormula(newFormula);
	//	}
	//
	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public void moveActiveFormula(ICalcIterator newActiveFormula) {
	//		calc_tree_.moveActiveFormula(newActiveFormula);
	//	}
	//
	/**
	 * Notify the Dialog Guide about interaction from the User
	 * 
	 * @param action
	 * @return true, if the action has been processed, false if the processing
	 *         the action has been denied
	 * @throws RemoteException
	 *  
	 */
	public boolean notifyUserAction(IUserAction action)
			throws DialogProtocolException, RemoteException {
		int request = action.getActionID();
		logger_.fatal(" WS->DG: notifyUserAction request= " + request);

		// two userActions dont care about the dialogphase
		//WN050420 would be nicer, if DialogGuide would set the Buttons
		switch (request) {
		case (UI_CALCULATE_1):
			if (phase_ == DIALOGPHASE_SPECIFY)
				request = UI_SPECIFY_TO_SOLVE;//WN#####################
			else
				request = UI_SOLVE_CALCULATE_1;
			break;
		case (UI_CALCULATE_ALL):
			if (phase_ == DIALOGPHASE_SPECIFY)
				request = UI_SPECIFY_CALCULATE_ALL;
			else
				request = UI_SOLVE_CALCULATE_ALL;
		}
		// all other userActions belong to a certain dialogphase

		switch (phase_) {
		case DIALOGPHASE_IDLE:
			throw new DialogProtocolException(request, phase_);
		case DIALOGPHASE_SPECIFY:
			if (request < UI_DUMMY_FIRST_SPECIFY
					|| request > UI_DUMMY_LAST_SPECIFY)
				throw new DialogProtocolException(request, phase_);
			break;
		case DIALOGPHASE_SOLVE:
			if (request < UI_DUMMY_FIRST_SOLVE || request > UI_DUMMY_LAST_SOLVE)
				throw new DialogProtocolException(request, phase_);
			break;
		}

		//*********** handle IUserAction's **********************
		CalcHead calc_head = new CalcHead();//WN0502023?????????
		try {
			calc_head = (CalcHead) calc_tree_.iterator().getFormula();
		} catch (Exception e) {
			e.printStackTrace();
		}
		switch (request) {
		case UI_SPECIFY_TRY_MATCH:
			throw new DialogNotImplementedException(request, phase_);
		//			try {
		//				tryMatch(calc_head, ((UserActionOnCalcHeadCompoundID) action)
		//						.getObjectID());
		//			} catch (NotInSpecificationPhaseException e) {
		//				throw new DialogMathException(request, phase_, e);
		//			}
		//			break;

		case UI_SPECIFY_TRY_REFINE:
			throw new DialogNotImplementedException(request, phase_);
		//			try {
		//				tryRefine(calc_head, ((UserActionOnCalcHeadCompoundID) action)
		//						.getObjectID());
		//			} catch (NotInSpecificationPhaseException e) {
		//				throw new DialogMathException(request, phase_, e);
		//			}
		//			break;

		case UI_SPECIFY_CHANGE_VIEW:
			calc_head.setViewStyle(((UserActionOnInt) action).getInt());
			break;

		case UI_SPECIFY_COMPLETE_CALCHEAD:
			calc_tree_.completeCalcHead();
			// old-style filling the fields of the calc-head by setting
			// the HELP_ME flag
			// calc_head_.setCalcHeadStatus(CalcHead.MODEL_ITEM_HELP_ME);
			// modifyCalcHead(calc_head_);
			break;
		case UI_SPECIFY_COMPLETE_METHOD:
		case UI_SPECIFY_COMPLETE_THEORY:
		case UI_SPECIFY_COMPLETE_PROBLEM:
		case UI_SPECIFY_COMPLETE_GIVEN:
		case UI_SPECIFY_COMPLETE_FIND:
		case UI_SPECIFY_COMPLETE_RELATE:
			throw new DialogNotImplementedException(request, phase_);

		case UI_SPECIFY_CHECK_CALCHEAD:
			calc_tree_.modifyCalcHead(calc_head);
			break;

		// was: startCalculation() WN050428 ? really ?
		case UI_SPECIFY_TO_SOLVE:
		case UI_SPECIFY_CALCULATE_ALL:
			//if (user_settings_.getValue(""))
			if (true)
				calc_tree_.completeCalcHead();
			else
				calc_tree_.modifyCalcHead(calc_head);

			if (calc_head.getCalcHeadStatus() != CalcHead.MODEL_ITEM_CORRECT)
				return false;

			try {
				calc_tree_.startSolving();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			// calc_tree_ = MathEngine.getMathEngine().startSolving(calcHead);
			// FIXME: AK020224: do not register with the calc_tree_ if finished
			// specifying a SUBPROBLEM
			//calc_tree_.addDataChangeListener(this);//WN050420-->startCalculation

			phase_ = DIALOGPHASE_SOLVE;//FIXXXME.WN050428
		// fall through to calculate

		case UI_SOLVE_CALCULATE_1:
		//WN050420 why not at (*) here ?
		case UI_SOLVE_CALCULATE_ALL:
			switch (request) {
			case UI_SPECIFY_CALCULATE_ALL:
			case UI_SOLVE_CALCULATE_ALL:
				calc_tree_.autoCalculate(IToCalc.SCOPE_CALCULATION, 0);
				break;
			case UI_SPECIFY_TO_SOLVE://WN050420 ???
			// break;
			default:
				calc_tree_.autoCalculate(IToCalc.SCOPE_CALCULATION, 1);
				//why not at (*) above ?
				break;
			}
			break;

		case UI_SOLVE_CALCULATE_SUBPROBLEM:
			calc_tree_.autoCalculate(IToCalc.SCOPE_SUBPROBLEM, 0);
			break;

		case UI_SOLVE_EDIT_ACTIVE_FORMULA:
			ui_control_listener_.doUIAction(new UserAction(UI_DO_EDIT_FORMULA));
			break;

		case UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE:
			CalcFormula formula = (CalcFormula) ((UserActionOnCalcElement) action)
					.getCalcElement();
			// if the position is null, the formula did not exist, so we try to
			// append a new formula
			if (formula.getPosition() == null) {
				calc_tree_.appendFormula(formula);
			} else {
				calc_tree_.replaceFormula(formula);
			}
			break;

		case UI_SOLVE_APPEND_USER_FORMULA:
			// appendFormula((Formula) ((UserActionOnCalcElement)
			// action).getCalcElement());
			ui_control_listener_
					.doUIAction(new UserAction(UI_DO_APPEND_FORMULA));
			break;

		case UI_SOLVE_MOVE_ACTIVE_FORMULA:
			calc_tree_.moveActiveFormula(((UserActionOnIterator) action)
					.getPosition());
			break;

		case UI_SOLVE_GET_PROPOSED_TACTIC:
			throw new DialogNotImplementedException(request, phase_);

		case UI_SOLVE_GET_APPLICABLE_TACTICS:
			throw new DialogNotImplementedException(request, phase_);

		case UI_SOLVE_SET_NEXT_TACTIC:
			calc_tree_
					.setNextTactic((Tactic) ((UserActionOnCalcElement) action)
							.getCalcElement());
			//WN041126 extracted from isac.bridge.CalcTree#setNextTactic @@@
			//autoCalculate(SCOPE_CALCULATION, 1);
			break;

		case UI_SOLVE_HELP_ENTERING_FORMULA:
			throw new DialogNotImplementedException(request, phase_);

		case UI_SOLVE_SHOW_ASSUMPTIONS:
			throw new DialogNotImplementedException(request, phase_);

		case UI_SOLVE_SHOW_DETAILS:
			throw new DialogNotImplementedException(request, phase_);

		default:
			throw new DialogUnknownActionException(request, phase_);
		}
		return true;
	}

	private void rmiBind() {
		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());
		}

		try {
			LocateRegistry.createRegistry(1099);
		} catch (java.rmi.RemoteException exc2) {
			System.err.println("can not create registry: " + exc2.getMessage());
		}

		String name = "//localhost/isac-DialogGuide";
		try {
			System.out.println("try to bind as " + name);
			Naming.rebind(name, this);
			System.out.println("Dialog Guide bound to " + name);
			//WN040906 was Object Manager ...
		} catch (java.rmi.ConnectException e) {
			System.err.println("failed to contact as " + name
					+ " (creating RMI-Server on localhost: 1099)");
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	//	public static void main(String[] args) {
	//		try {
	//			new DialogGuide(args[0]);
	//
	//		} catch (RemoteException e) {
	//			e.printStackTrace();
	//		}
	//	}

	//	/**
	//	 * @see
	// isac.util.interfaces.IToCalc#completeCalcHead(isac.util.formulae.CalcHead)
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public void completeCalcHead(CalcHead calcHead) {
	//		// TODO Auto-generated method stub
	//
	//	}
	//
	//	/**
	//	 * @see
	// isac.util.interfaces.IToCalc#completeCalcHead(isac.util.formulae.CalcHead,
	//	 * int)
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public void completeCalcHead(CalcHead calcHead, int completeItem) {
	//		// TODO Auto-generated method stub
	//
	//	}
	//
	//	/**
	//	 * (non-Javadoc)
	//	 *
	//	 * @see isac.util.interfaces.IToCalc#completeCalcHead()
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public void completeCalcHead() throws RemoteException {
	//		// TODO Auto-generated method stub WN050223
	//
	//	}
	//
	//	/**
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 *
	//	 * @see
	// isac.util.interfaces.IToCalc#intermediateSteps(isac.util.interfaces.ICalcIterator)
	//	 */
	//	public int intermediateSteps(ICalcIterator i) throws RemoteException {
	//		// TODO Auto-generated method stub
	//		return 0;
	//	}
	//
	//	/**
	//	 * @see
	// isac.util.interfaces.IToCalc#getElementsFromTo(isac.util.interfaces.ICalcIterator,
	//	 * isac.util.interfaces.ICalcIterator, java.lang.Integer, boolean)
	//	 * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
	//	 */
	//	public Vector getElementsFromTo(ICalcIterator iterator_from,
	//			ICalcIterator iterator_to, Integer level,
	//			boolean result_includes_tactics) throws RemoteException {
	//		// TODO Auto-generated method stub
	//		return null;
	//	}

}